module Configuration where

import SymbVal
import Datapath

import List
import Monad

type Pair = (NodeName, SymbVal)
type Row = [SymbVal]
type Column = (NodeName, [SymbVal])
type Table = [Row]
type Config = ([NodeName], Table)

type ConfigInfo = [(NodeName, [NodeName])]

flat :: Config
flat = ([], [])

--Table Functions

--attempts to combine two tables without increasing their length.
--Uses the open value
combineTables :: Table -> Table -> Maybe Table
combineTables c1s c2s =
 do
   let r1 = length (head c1s)
   let r2 = length (head c2s)
   when (r1 /= r2) Nothing
   let l1 = length c1s
   let l2 = length c2s
   let c1s' = if l2 > l1 then c1s ++ (blankTable (l2 - l1, r1)) else c1s
   let c2s' = if l1 > l2 then c2s ++ (blankTable (l1 - l2, r2)) else c2s
   let zips = zip c1s' c2s'
   rs <- mapM (uncurry combineRows) zips
   return rs

combineRows :: Row -> Row -> Maybe Row
combineRows xs ys = 
 do
   let zips = zip xs ys
   r <- mapM (uncurry combineM) zips
   return r

--Combines tables by pushing one down until it fits
addTables :: Table -> Table -> Table
addTables x y = case combineTables x y of
   Just z -> z
   Nothing -> addTables x (pushDown y)

--Adds tables and returns how many rows it was pushed down
addTablesN :: Table -> Table -> (Table, Int)
addTablesN x y = addTablesN' x y 0
 where
  addTablesN' x y n = case combineTables x y of
   Just z -> (z, n)
   Nothing -> addTablesN' x (pushDown y) (n+1)

--Repeats a table a number of times equal to its length
--and returns how many rows larger the result is.
{- IE  1                 1
        2          ==>   12  
	 3               123   & 3
	                 123
			  23
			   3

-}
repTableL :: Table -> (Table, Int)
repTableL x = repTableL' x x (length x)
 where
  repTableL' x y 1 = addTablesN x y
  repTableL' x y n = repTableL' (addTables x y) y (n-1)
   where
    
--Performs concatenation and overlapping scheduling
scheduleTable :: Table -> Table
scheduleTable x = scheduleTable' x x
 where
  scheduleTable' x y = if n < ((length y) - 2) 
                          then scheduleTable' x' y
                          else x
   where
    (x', n) = addTablesN x y

--Schedules a table and then compresses the columns.
{-   Original     Repeated  Scheduled  Compressed
        1           1
	 2    ==>   12   ==>  123   ==> 123
	1 3         123       143        4
	 4          143
	             4
	 
-}
optimize :: Config -> [Column]
optimize (ns, t) = final
 where
  sched = scheduleTable t
  (t', n1) = repTableL sched
  (t'',n2) = addTablesN t' sched
  mid = drop n1 (take n2 t'')
  final = map compressCol (configToCols (ns, mid))
  
--Compresses columns
compressCol :: Column -> Column
compressCol (n, (c:cs)) = case filter (c /=) cs of
                            [] -> (n,[c]) --could search for longer patterns later
                            z -> (n, (c:cs))

--Creates a blank table of the given dimensons
blankTable :: (Int, Int) -> Table
blankTable (x,y) = replicate x (blankRow y)

blankRow :: Int -> Row
blankRow y = replicate y Unknown

pushDown :: Table -> Table
pushDown [] = []
pushDown (x:xs) = (blankRow (length x)):x:xs

tableWidth :: Table -> Int
tableWidth [] = 0
tableWidth (x:xs) = length x

dupTable :: Table -> Table
dupTable t = addTables t t

repeatTable :: Int -> Table -> Table
repeatTable 0 t = []
repeatTable 1 t = t
repeatTable k t = addTables (repeatTable (k-1) t) t

--Configuration Functions

configToCols :: Config -> [Column]
configToCols (ns, rs) = zip ns (transpose rs)

colsToConfig :: [Column] -> Config
colsToConfig cs = (headings, rows)
 where
   headings = map fst cs
   rows = transpose (map snd cs)

pairsToCols :: [[Pair]] -> [Column]
pairsToCols [] = []
pairsToCols pss = map (makeCol pss) (getVarRows pss [])
 where
  getVarRows [] vs = vs
  getVarRows (ps:pss) vs = getVarRows pss (getVars ps vs)
  getVars [] vs = vs
  getVars ((p,v):ps) vs = if elem p vs 
                      then getVars ps vs
		      else getVars ps (p:vs)
  makeCol ps v = (v, map makeCol' (map (lookup v) ps))
  makeCol' (Just x) = x
  makeCol' Nothing = Unknown
  
concatConfig :: Config -> Config -> Config
concatConfig (ns1, bs1) (ns2, bs2) = (ns1 ++ ns2, bs')
 where
  bs' = map (uncurry (++)) zips
  zips = zip bs1 bs2

concatConfigs :: [Config] -> Config
concatConfigs [] = ([],[])
concatConfigs [x] = x
concatConfigs (x:xs) = foldl concatConfig x xs

--gives the increase in size from adding two tables
sizeIncrease :: Table -> Table -> Int
sizeIncrease t1 t2 = snd (addTablesN t1 t2)
{- If weird things happen go back to this!!
sizeIncrease x y = (length res) - larger
 where
  res = addTables x y
  larger = if (length x) > (length y) then length x else length y
-}
  
prefixVal :: String -> SymbVal -> SymbVal
prefixVal p (Var s) = Var (p ++ ":" ++ s)
prefixVal p (VarIx s n) = VarIx (p ++ ":" ++ s) n
prefixVal p z = z

subConfig :: Config -> NodeName -> Config
subConfig (ns, bs) m = (ns', bs')
 where
  ns' = map (\x -> m ++ ":" ++ x) ns
  bs' = map (map (prefixVal m)) bs

subConfigs cf ms = concatConfigs cfs
 where
  cfs = map (subConfig cf) ms

--Returns all exposed join points and their associated choices
configInfo :: Datapath -> ConfigInfo
configInfo (ns, es) = zips
 where
  jps = map nodename (filter isExposedJoinPoint ns)
  prvs = map getPrevs jps
  getPrevs jp = map fst (filter (isNext (jp, 0)) es)
  zips = zip jps prvs
  isExposedJoinPoint n = (isJoinPoint n) && ((filter (isNext (nodename n, 1)) es) == [])

--Returns all exposed join points
configChoices :: Datapath -> [NodeName]
configChoices dp = map fst (configInfo dp)

--Creates a blank config of length n for a given dp
blankConfig :: Datapath -> Int -> Config
blankConfig (ns, es) n = (jps, blank)
 where
  jps = map nodename (filter isJoinPoint ns)
  blank = blankTable (n,length jps)

-- IO Functions
  
displayCols :: [Column] -> IO ()
displayCols cs = displayConfig (colsToConfig cs)

displayConfig :: Config -> IO ()
displayConfig (headings, bt) = 
 do
   let h = concat (intersperse "\t" headings)
   putStrLn h
   putStrLn (replicate 80 '=')
   mapM printRow bt
   return ()
 where
  printRow r = putStrLn $ concat (intersperse "\t" (map show r))

displayConfigN :: Int -> Config -> IO ()
displayConfigN n (headings, bt) = 
 do
   let h = concat (intersperse "\t" headings)
   putStr "CC\t|\t"
   putStrLn h
   putStr "--------+"
   putStrLn (replicate (8 * ((length headings)+1)) '-')
   printRows bt 0 n
   return ()
 where
  printRows [] _ _ = return ()
  printRows _ _ 0 = return ()
  printRows (r:rs) c n = 
   do
    putStr (show c)
    putStr "\t|\t"
    putStrLn $ concat (intersperse "\t" (map show r))
    printRows rs (c+1) (n-1)
