module Run where

import Datapath
import Configuration
import SymbVal
import UsagePattern

import STATE
import ErrM

import List

type DelayState = [Pair]

--State during simulation
data DPState = DPState
 {
   datapath    :: Datapath,
   last_cc     :: Int,
   cur_cc      :: Int,
   cur_st      :: DelayState,
   next_st     :: DelayState,
   next_up      :: ([Pair], [Pair]),
   up          :: UsagePattern,
   st          :: [DelayState]
 } deriving (Show, Eq)

type RunState = STATE DPState

currentCC :: RunState Int
currentCC =
 do 
   st <- readSTATE
   return (cur_cc st)

currentDP :: RunState Datapath
currentDP =
 do 
   st <- readSTATE
   return (datapath st)

prevNode :: InputPort -> RunState Node
prevNode p =
 do
   dp <- currentDP
   return (lookupNode dp (getPrevNode_n dp p))

addUseIn :: Pair -> RunState ()
addUseIn p =
 do
   st <- readSTATE
   let (ins, outs) = next_up st
   if elem p ins
    then return ()
    else do
      writeSTATE st {next_up = (p : ins, outs)}
      return ()

addUseOut :: Pair -> RunState ()
addUseOut p =
 do
   st <- readSTATE
   let (ins, outs) = next_up st
   writeSTATE st {next_up = (ins, p : outs)}
   return ()
   
nextVal :: NodeName -> RunState (Maybe SymbVal)
nextVal n =
 do
   st <- readSTATE
   case lookup n (next_st st) of
     Just x -> return (Just x)
     Nothing -> do --this stops infinite loops on delays
       updateNextVal (n, Unknown)
       return Nothing

updateNextVal :: Pair -> RunState ()
updateNextVal (n, v) = 
 do
   st <- readSTATE
   case lookup n (next_st st) of
     Just x -> 
       writeSTATE st {next_st = (n,v) : ((next_st st) \\ [(n,x)])} 
     Nothing -> do
       writeSTATE st {next_st = (n,v) : (next_st st)} 
       return ()
   
curVal :: NodeName -> RunState SymbVal
curVal n =
 do
   st <- readSTATE
   case lookup n (cur_st st) of
     Nothing -> fail ("Missing state for node " ++ n)
     Just c -> return c
   
drive :: Node -> RunState SymbVal
drive (Input i) =
 do
   cc <- currentCC
   let v = VarIx i cc
   addUseIn (i, v)
   return v
drive (Output o) =
 do
   pred <- prevNode (o,0)
   v <- drive pred
   addUseOut (o, v)
   return v
drive (Delay d i) =
 do
   next <- nextVal d
   case next of
     Just x -> curVal d
     Nothing -> do --calculate new value for next CC
       pred <- prevNode (d, 0)
       v <- drive pred
       updateNextVal (d, v)
       curVal d
drive (Function f fun) =
 do
   dp <- currentDP
   let k = nParams fun
   let prvs = getPrevNodes dp f
   let pns = map (lookupNode dp) prvs
   vals <- mapM drive pns
   return (symbApply fun vals)
{- This version is hard for humans to use
drive (JoinPoint j) =
 do
   con <- prevNode (j, 2)
   c <- drive con
   case c of
     Unknown -> return Unknown
     (VBool False) -> do
       pred <- prevNode (j,0)
       return (drive pred)
     (VBool True) -> do
       pred <- prevNode (j,1)
       return (drive pred)
     z -> return Unknown -}
drive (JoinPoint j) =
 do
   con <- prevNode (j, 1)
   c <- drive con
   case c of
     Unknown -> return Unknown
     (Var s) -> do
       dp <- currentDP
       let pns = getPrevNodes_n dp (j,0)
       if elem s pns then do
         let pn = lookupNode dp s
         drive pn
        else fail ("Node " ++ s ++ " is not connected to Join Point " ++ j)
     z -> return Unknown

semCC :: RunState ()
semCC =
 do
   st <- readSTATE
   let dp = datapath st
   let delays = getDelays dp
   let outputs = getOutputs dp
   mapM_ drive (outputs ++ delays)
   st' <- readSTATE
   let (ins, outs) = up st'
   let (is, os) = next_up st'
   let up' = (is : ins, os : outs)
   writeSTATE st' {cur_cc = (cur_cc st) + 1, next_up = ([],[]), up = up', cur_st = next_st st', next_st = []}
   

graphsem :: RunState UsagePattern
graphsem = 
 do
  st <- readSTATE
  if (cur_cc st) <= (last_cc st)
     then do
        semCC
	graphsem
     else do
       let (ins, outs) = up st
       return (reverse ins, reverse outs)

run :: Datapath -> Config -> Int -> UsagePattern
run dp cf k = 
 case appSTATE graphsem newstate of
    Ok (x,y) -> x
  where
   newstate = DPState dp' k 0 initial [] ([],[]) ([],[]) []
   dp' = case appSTATE (addController cf) dp of
     Ok (x,y) -> x
   initial = getInitialState dp'

runIO :: Datapath -> Config -> Int -> IO ()
runIO dp c n=
 do
   let res = run dp c n
   displayUP_n (n+1) res

runHTML :: String -> Datapath -> Config -> Int -> IO ()
runHTML file dp c n =
 do 
   let res = run dp c n
   let html = upToHTML_n (n+1) res
   writeFile file html

col2Delays :: Column -> Result ()
col2Delays (nm, vs) =
 do
  let k = (length vs) - 1
  let nm' = "control:" ++ nm ++ "_d"
  constantList vs nm'

addController :: Config -> Result Datapath
addController (ns, bs) =
 do
   let k = length ns
   let cols = optimize (ns, bs)
   mapM col2Delays cols
   let heads = map (\x -> "control:" ++ x ++ "_d_0") ns
   let zips = zip heads (zip ns (replicate k 1))
   mapM connect_n zips
   dp <- readSTATE
   return dp
{-
testConfig :: Int -> Config -> Result [Column]
testConfig k cf =
 do
   addController cf
   vals <- mapM stepDP (range (0, k))
   let regs = getDelays vals
   let vs = map (getVals vals) regs
   return vs

getDelays :: [[(NodeName, SymbVal)]] -> [NodeName]
getDelays ds = getDelays' [] (head ds)
 where
  getDelays' rs [] = rs
  getDelays' rs ((r,v):vs) = if elem r rs then getDelays' rs vs else getDelays' (r:rs) vs

getVals :: [[(NodeName, SymbVal)]] -> NodeName -> Column
getVals vals n = (n, concatMap (getVals' n) vals)
 where
  getVals' n vs = concatMap (getVals'' n) vs
  getVals'' n (nm,v) = if n == nm then [v] else [] 

runDot :: Datapath -> Config -> Int -> IO ()
runDot dp cf n =
 do
   let cols = run dp cf n
   let config = colsToConfig cols
   let gs = map (svToDot . last . snd) cols
   let zips = zip (fst config) gs
   displayConfigN (n+1) config
   mapM_ writeDotFile zips
-}      
test dp f = 
  case appSTATE f dp of
    (Ok a) -> case a of
      (x,y) -> do
        putStr "Result:"
        putStrLn (show x)
        putStrLn "State:"
        putStrLn (show y)

